@@ -151,6 +151,7 @@ class Attention(nn.Module):
     def __init__(
         self,
         dim,
+        dim_context = None,
         heads = 4,
         dim_head = 32,
         norm = False,
@@ -159,6 +160,7 @@ class Attention(nn.Module):
     ):
         super().__init__()
         hidden_dim = dim_head * heads
+        dim_context = default(dim_context, dim)
 
         self.time_cond = None
 
@@ -176,10 +178,10 @@ class Attention(nn.Module):
         self.heads = heads
 
         self.norm = LayerNorm(dim) if norm else nn.Identity()
-        self.norm_context = LayerNorm(dim) if norm_context else nn.Identity()
+        self.norm_context = LayerNorm(dim_context) if norm_context else nn.Identity()
 
         self.to_q = nn.Linear(dim, hidden_dim, bias = False)
-        self.to_kv = nn.Linear(dim, hidden_dim * 2, bias = False)
+        self.to_kv = nn.Linear(dim_context, hidden_dim * 2, bias = False)
         self.to_out = nn.Linear(hidden_dim, dim, bias = False)
 
     def forward(
@@ -193,6 +195,9 @@ class Attention(nn.Module):
 
         context = default(context, x)
 
+        if x.shape[-1] != self.norm.gamma.shape[-1]:
+            print(context.shape, x.shape, self.norm.gamma.shape)
+
         x = self.norm(x)
 
         if exists(self.time_cond):
@@ -215,6 +220,22 @@ class Attention(nn.Module):
         out = rearrange(out, 'b h n d -> b n (h d)')
         return self.to_out(out)
 
+class PEG(nn.Module):
+    def __init__(
+        self,
+        dim
+    ):
+        super().__init__()
+        self.ds_conv = nn.Conv2d(dim, dim, 3, padding = 1, groups = dim)
+
+    def forward(self, x):
+        b, n, d = x.shape
+        hw = int(math.sqrt(n))
+        x = rearrange(x, 'b (h w) d -> b d h w', h = hw)
+        x = self.ds_conv(x)
+        x = rearrange(x, 'b d h w -> b (h w) d')
+        return x
+
 class FeedForward(nn.Module):
     def __init__(self, dim, mult = 4, time_cond_dim = None):
         super().__init__()
@@ -251,6 +272,63 @@ class FeedForward(nn.Module):
 
 # model
 
+class RINBlock(nn.Module):
+    def __init__(
+        self,
+        dim,
+        latent_self_attn_depth,
+        dim_latent = None,
+        final_norm = True,
+        **attn_kwargs
+    ):
+        super().__init__()
+        dim_latent = default(dim_latent, dim)
+
+        self.latents_attend_to_patches = Attention(dim_latent, dim_context = dim, norm = True, norm_context = True, **attn_kwargs)
+
+        self.latent_self_attns = nn.ModuleList([])
+        for _ in range(latent_self_attn_depth):
+            self.latent_self_attns.append(nn.ModuleList([
+                Attention(dim_latent, norm = True, **attn_kwargs),
+                FeedForward(dim_latent)
+            ]))
+
+        self.latent_final_norm = LayerNorm(dim_latent) if final_norm else nn.Identity()
+
+        self.patches_peg = PEG(dim)
+        self.patches_self_attn = LinearAttention(dim, norm = True, **attn_kwargs)
+        self.patches_self_attn_ff = FeedForward(dim)
+
+        self.patches_attend_to_latents = Attention(dim, dim_context = dim_latent, norm = True, norm_context = True, **attn_kwargs)
+        self.patches_cross_attn_ff = FeedForward(dim)
+
+    def forward(self, patches, latents, t):
+        patches = self.patches_peg(patches) + patches
+
+        # latents extract or cluster information from the patches
+
+        latents = self.latents_attend_to_patches(latents, patches, time = t) + latents
+
+        # latent self attention
+
+        for attn, ff in self.latent_self_attns:
+            latents = attn(latents, time = t) + latents
+            latents = ff(latents, time = t) + latents
+
+        # additional patches self attention with linear attention
+
+        patches = self.patches_self_attn(patches, time = t) + patches
+        patches = self.patches_self_attn_ff(patches) + patches
+
+        # patches attend to the latents
+
+        patches = self.patches_attend_to_latents(patches, latents, time = t) + patches
+
+        patches = self.patches_cross_attn_ff(patches, time = t) + patches
+
+        latents = self.latent_final_norm(latents)
+        return patches, latents
+
 class RIN(nn.Module):
     def __init__(
         self,
@@ -258,14 +336,16 @@ class RIN(nn.Module):
         image_size,
         patch_size = 16,
         channels = 3,
-        depth = 6,                      # weight tied depth. weight tied layers basically is recurrent, with latents as hiddens
+        depth = 6,                      # number of RIN blocks
         latent_self_attn_depth = 2,     # how many self attentions for the latent per each round of cross attending from pixel space to latents and back
+        dim_latent = None,              # will default to image dim (dim)
         num_latents = 256,              # they still had to use a fair amount of latents for good results (256), in line with the Perceiver line of papers from Deepmind
         learned_sinusoidal_dim = 16,
         **attn_kwargs
     ):
         super().__init__()
         assert divisible_by(image_size, patch_size)
+        dim_latent = default(dim_latent, dim)
 
         self.channels = channels # times 2 due to self-conditioning
 
@@ -293,7 +373,7 @@ class RIN(nn.Module):
             nn.Linear(pixel_patch_dim * 2, dim)
         )
 
-        self.pos_emb = nn.Parameter(torch.randn(num_patches, dim))
+        self.axial_pos_emb = nn.Parameter(torch.randn(2, patch_height_width, dim) * 0.02)
 
         self.to_pixels = nn.Sequential(
             LayerNorm(dim),
@@ -301,35 +381,21 @@ class RIN(nn.Module):
             Rearrange('b (h w) (c p1 p2) -> b c (h p1) (w p2)', p1 = patch_size, p2 = patch_size, h = patch_height_width)
         )
 
-        self.latents = nn.Parameter(torch.randn(num_latents, dim))
+        self.latents = nn.Parameter(torch.randn(num_latents, dim_latent))
+        nn.init.normal_(self.latents, std = 0.02)
 
         self.init_self_cond_latents = nn.Sequential(
-            FeedForward(dim),
-            LayerNorm(dim)
+            FeedForward(dim_latent),
+            LayerNorm(dim_latent)
         )
 
         nn.init.zeros_(self.init_self_cond_latents[-1].gamma)
 
         # the main RIN body parameters  - another attention is all you need moment
 
-        self.depth = depth
-
         attn_kwargs = {**attn_kwargs, 'time_cond_dim': time_dim}
 
-        self.latents_attend_to_patches = Attention(dim, norm = True, norm_context = True, **attn_kwargs)
-
-        self.latent_self_attns = nn.ModuleList([])
-        for _ in range(latent_self_attn_depth):
-            self.latent_self_attns.append(nn.ModuleList([
-                Attention(dim, norm = True, **attn_kwargs),
-                FeedForward(dim)
-            ]))
-
-        self.patches_self_attn = LinearAttention(dim, norm = True, **attn_kwargs)
-        self.patches_self_attn_ff = FeedForward(dim)
-
-        self.patches_attend_to_latents = Attention(dim, norm = True, norm_context = True, **attn_kwargs)
-        self.patches_cross_attn_ff = FeedForward(dim)
+        self.blocks = nn.ModuleList([RINBlock(dim, dim_latent = dim_latent, latent_self_attn_depth = latent_self_attn_depth, **attn_kwargs) for _ in range(depth)])
 
     def forward(
         self,
@@ -362,31 +428,14 @@ class RIN(nn.Module):
 
         patches = self.to_patches(x)
 
-        patches = patches + self.pos_emb
+        pos_emb_h, pos_emb_w = self.axial_pos_emb
+        pos_emb = rearrange(pos_emb_h, 'i d -> i 1 d') + rearrange(pos_emb_w, 'j d -> 1 j d')
+        patches = patches + rearrange(pos_emb, 'i j d -> (i j) d')
 
         # the recurrent interface network body
 
-        for _ in range(self.depth):
-            # latents extract or cluster information from the patches
-
-            latents = self.latents_attend_to_patches(latents, patches, time = t) + latents
-
-            # latent self attention
-
-            for attn, ff in self.latent_self_attns:
-                latents = attn(latents, time = t) + latents
-                latents = ff(latents, time = t) + latents
-
-            # additional patches self attention with linear attention
-
-            patches = self.patches_self_attn(patches, time = t) + patches
-            patches = self.patches_self_attn_ff(patches) + patches
-
-            # patches attend to the latents
-
-            patches = self.latents_attend_to_patches(patches, latents, time = t) + patches
-
-            patches = self.patches_cross_attn_ff(patches, time = t) + patches
+        for block in self.blocks:
+            patches, latents = block(patches, latents, t)
 
         # to pixels
 
